#include <cassert>
#include <cstdlib>
#include <iostream>
#include <string>
#include "Spreadsheet.h"
#include "Function.h"

// Test driver and main loop.

using namespace std;

void testCell()
{
    // A basic test which works even if
    // Spreadsheet and Function aren't
    // implemented yet:
    Cell c1;
    c1.setValue(42.89);
    Cell c2;
    c1.subscribe(&c2);
    c1.setValue(19); // Should call update()
    c1.printOn(cout);
}

void testSpreadsheet()
{
    // Test the spreadsheet without
    // functions or dependencies:

    Spreadsheet s;
    s.setSize(3,3);
    s.at(1,1).setValue(42.6);
    s.at(1,2).setValue(13.9);
    s.printOn(cout);
}

void testAll()
{
    // Test the spreadsheet with
    // functions enabled:

    const Value v1 = 42.6;
    const Value v2 = 58.23;
    const Value v3 = 23.1;

    Spreadsheet s;
    s.setSize(3,3);
    s.at(1,1).setValue(v1);
    s.at(1,2).setValue(v2);

    // When we set a value, it should
    // stay set:
    assert(s.at(1,1).getValue() == v1);
    assert(s.at(1,2).getValue() == v2);

    BinOp *f1 = new BinOp(add, &s.at(1,1), &s.at(1,2));
    s.at(2,1).setFunc(f1);
    s.printOn(cout);

    // Check that adding two numbers works:
    assert(s.at(2,1).getValue() == v1+v2);

    BinOp *f2 = new BinOp(multiply, &s.at(2,1), &s.at(1,2));

    s.at(2,2).setFunc(f2);
    s.printOn(cout);

    // Check that multiplication works:
    assert(s.at(2,2).getValue() == (v1 + v2) * v2);

    s.at(1,1).setValue(v3);
    s.printOn(cout);

    // Check that notification and value updates work:
    assert(s.at(2,2).getValue() == (v3 + v2) * v2);
}

void initSpreadsheet(Spreadsheet &s)
{
    // Set up the formula:
    // ((a+b) / (a*b)) - ((a+b) % (a*b))
    //
    // Cell assignments:
    //  1,1 : a
    //  1,2 : b
    //  2,1 : a+b
    //  2,2 : a*b
    //  3,1 : (a+b)/(a*b)
    //  3,2 : (a+b)%(a*b)
    //  4,1 : ((a+b) / (a*b)) - ((a+b) % (a*b))
    
    s.setSize(4,2);
    s.at(2,1).setFunc(new BinOp(add, &s.at(1,1), &s.at(1,2))); 
    s.at(2,2).setFunc(new BinOp(multiply, &s.at(1,1), &s.at(1,2))); 
    s.at(3,1).setFunc(new BinOp(divide, &s.at(2,1), &s.at(2,2)));
    s.at(3,2).setFunc(new BinOp(modfunc, &s.at(2,1), &s.at(2,2)));
    s.at(4,1).setFunc(new BinOp(subtract, &s.at(3,1), &s.at(3,2)));
}

bool doCmd(Spreadsheet &s)
{
    // Execute one command and display the spreadsheet.
    cout << "->";

    string cmd;
    cin >> cmd;
    if (cmd == "set")
    {
        int row,col;
        Value val;
        cin >> row;
        cin >> col;
        cin >> val;

        s.at(row,col).setValue(val);

        cout << "--------------------" << endl;
        s.printOn(cout);
        cout << endl;
    }
    else if (cmd == "quit")
        return false; // Exit the program
    else
    {
        cerr << "Unknown command: " << cmd << endl;
        cerr << "Available commands:\nset row, col, value\nquit" << endl;
    }
    return true; // Keep accepting commands
}


void run()
{
    // Initialize the spreadsheet to a hard-coded
    // function and enter an input loop to set
    // the values:
    Spreadsheet s;
    initSpreadsheet(s);
    s.at(1,1).setValue(0);
    s.at(1,2).setValue(0);

    while (doCmd(s)) {}
}


main()
{
    // Following tests can be re-enabled
    // for regression testing at any time:

    //testCell();
    //testSpreadsheet();
    //testAll();

    run();
}
