#include <iostream>
#include <iomanip>
#include <limits>
#include "Spreadsheet.h"
#include "Function.h"
using namespace std;

const Value Cell::NULL_VALUE = numeric_limits<Value>::infinity() ;

Cell::Cell() 
: cachedValue(NULL_VALUE), func(NULL)
{}

Cell::~Cell()
{
    clear();
}

void Cell::update(const Publisher *)
{
    // If the cell has a function,
    // evaluate it and update the
    // cell's value:
    if (func)
    {
        cachedValue = func->evaluate();
        notify();
    }
}

void Cell::setValue(const Value &v)
{
    // Clear the cell of any associated
    // function, set the cached value,
    // and notify dependents about
    // the change:
    clear();
    cachedValue = v;
    notify();
}

void Cell::setFunc(IFunction *f)
{
    // Clear any existing function, set the
    // new function, and update the cached
    // value:
    clear();
    func = f;

    // Subscribe to the sources of the function:
    list<Cell*> srcs = func->getSources();
    for(list<Cell*>::iterator i = srcs.begin();
        i!=srcs.end();
        ++i)
        {
            (**i).subscribe(this);
        }
    
    // Force an immediate update:
    update(NULL);
}

void Cell::clear()
{
    // Delete any existing function and
    // set the object to a null state:
    if (func)
    {
        // Unsubscribe this cell from its
        // current publishers:
        list<Cell*> srcs = func->getSources();
        for(list<Cell*>::iterator i = srcs.begin();
            i!=srcs.end();
            ++i)
            {
                (**i).unsubscribe(this);
            }
        delete func;
        func = NULL;
    }
    cachedValue = NULL_VALUE;
}

bool Cell::isNull() const
{
    return (!func && (cachedValue==NULL_VALUE));
}

void Cell::printOn(ostream &o) const
{
    // Print out the current value
    // if there is one, else print a
    // default null string:
    if (!isNull())
        o << cachedValue;
    else
        o << "";
}


//----------------------------------------------
Spreadsheet::~Spreadsheet()
{
    // Clear out the cells (resets inter-cell
    // references and unsubscribes cells from
    // each other; must do this before destroying
    // them)
    clear();
}


void Spreadsheet::setSize(int rows,int cols)
{
    matrix.clear();
    CellRow init(cols); // Create a row #cols long
    for(int i=0;i<rows;++i)
    {
        matrix.push_back(init); // Add a row
    }
}

void Spreadsheet::clear()
{
    for(Matrix::iterator row = matrix.begin();
        row != matrix.end();
        ++row)
        {
            for(CellRow::iterator i = row->begin();
                i != row->end();
                ++i)
                {
                    i->clear();
                }
        }
}

void Spreadsheet::printOn(ostream &o) const
{
    const int CELLWIDTH = 8;
    for(Matrix::const_iterator row = matrix.begin();
        row != matrix.end();
        ++row)
        {
            for(CellRow::const_iterator i = row->begin();
                i != row->end();
                ++i)
                {
                    o << setw(CELLWIDTH);
                    i->printOn(o);
                }
            o << endl;
        }
}

