#ifndef SPREADSHEET_H_
#define SPREADSHEET_H_

// Contains Spreadsheet and Cell classes.

#include <iosfwd>
#include <vector>
#include "Defs.h"
#include "PubSub.h"

class IFunction;

// A Cell is a container of a value.  It has
// a location within a Spreadsheet.  It is
// responsible for updating itself in response
// to changes in other Cells.
class Cell : public Publisher, public ISubscriber
{
public:
    Cell();
    virtual ~Cell();

    // Implementation of ISubscriber:
    virtual void update(Publisher const *p);

    // Cell operations:

    // Retrieve the most current value of the cell:
    Value getValue() const {return cachedValue;}

    // Set the cell's value (removes any current function)
    void setValue(Value const &v);

    // Set the cell's function (removes any current value)
    // (Note: The Cell takes ownership of the function object.)
    void setFunc(IFunction *);

    // Display the cell's contents
    void printOn(std::ostream &) const;

    // Clear the cell to a null state
    void clear();
    
    // Tell if the cell is in a null state
    bool isNull() const;    
private:
    static const Value NULL_VALUE; // Used to indicate null state
    
    Value cachedValue;      // Cached value of the cell
    IFunction *func;        // Associated function (if any)
};

// A Spreadsheet is a container of Cells.  It is responsible
// for creating and destroying Cells and displaying them
// in a 2D array.
class Spreadsheet
{
public:
    ~Spreadsheet();

    // Set the size of the spreadsheet (destroys current
    // values):
    void setSize(int rows,int cols);

    // Get access to a cell at (row,col); row, col are 
    // 1-based:
    Cell &at(int row, int col) {return matrix[row-1][col-1];}

    // Print the spreadsheet:
    void printOn(std::ostream &) const;

    // Clear the spreadsheet's functions and values:
    void clear();
private:
    typedef std::vector<Cell> CellRow;
    typedef std::vector<CellRow> Matrix;
    Matrix matrix;
};

#endif
